import importlib.util
# import sys
from hashlib import sha1
import datetime
import dateutil
import benedict
from beancount.core.number import D
from beancount.core import amount
import os
from .logger import logger


def load_config(path):

    # see https://stackoverflow.com/questions/67631/how-to-import-a-module-given-the-full-path
    module_name = "config"
    spec = importlib.util.spec_from_file_location(module_name, path)
    module = importlib.util.module_from_spec(spec)
    # sys.modules[module_name] = module
    spec.loader.exec_module(module)
    # print(module.__dict__)
    # print(module.__dict__.get("account"))
    # exit()
    return module.__dict__


def critical_error(*args):
    logger.critical(*args)
    exit(1)


def map_dict(value, data: dict):
    """
    >>> map_dict({'bc-payee': 'payee', 'date': ['payee', 'date', lambda test: 'bar', {'meh': 'payee'}]}, {'payee': 'foo-payee', 'date': 'foo-date'})
    {'bc-payee': 'foo-payee', 'date': ['foo-payee', 'foo-date', 'bar', {'meh': 'foo-payee'}]}
    """
    ret = {}

    if callable(value):
        return value(benedict.benedict(data))

    if isinstance(value, bytes):
        return value.decode()

    if isinstance(value, str):
        return benedict.benedict(data).get(value)

    if isinstance(value, set):
        return {map_dict(x, data) for x in value}

    if isinstance(value, list):
        return [map_dict(x, data) for x in value]

    if isinstance(value, dict):
        return {k: map_dict(v, data) for k, v in value.items()}

    return value


def hash_str(s: str):
    """
    >>> hash_str("foo")
    '0beec7b5ea3f0fdbc95d0dd47f3c5bc275da8a33'
    """
    # https://death.andgravity.com/stable-hashing
    hash_value = sha1(s.encode("utf-8")).hexdigest()
    return hash_value


def parse_date(thing) -> datetime.date:
    """
    Turn a random variable into a date

    - date: return unchanged
    - datetime -> datetime's date
    - string: gets parsed via dateutil

    >>> parse_date(datetime.date(2023,2,3))
    datetime.date(2023, 2, 3)
    >>> parse_date(datetime.datetime(2023,1,1,12,34,56))
    datetime.date(2023, 1, 1)
    >>> parse_date("2017-11-05T00:06:18")
    datetime.date(2017, 11, 5)
    """

    try:
        if isinstance(thing, datetime.datetime):
            return thing.date()

        if isinstance(thing, datetime.date):
            return thing

        return dateutil.parser.parse(thing).date()
    except Exception as e:
        raise type(e)("parse_date: %s" % str(e))


def parse_amount(thing, default_currency="EUR") -> amount.Amount:
    """
    Create a beancount amount object from any values

    >>> parse_amount([120, "USD"])
    120 USD
    >>> parse_amount("120")
    120 EUR
    >>> parse_amount(120)
    120 EUR
    >>> parse_amount("120", "GBP")
    120 GBP
    """
    try:
        if thing is None:
            return None

        if isinstance(thing, amount.Amount):
            return thing

        if isinstance(thing, list):
            return amount.Amount(D(thing[0]), thing[1])

        return amount.Amount(D(thing), default_currency)
    except Exception as e:
        raise type(e)("parse_amount: %s" % str(e))


def path_to_dict(path):

    def subpath_to_dict(subpath, d={}):

        name = os.path.basename(subpath)

        if os.path.isdir(subpath):
            if name not in d:
                d[name] = {'_': []}
            for x in os.listdir(subpath):
                subpath_to_dict(os.path.join(subpath,x), d[name])
        else:
            d['_'].append(benedict.benedict(subpath))
        return d

    return list(subpath_to_dict(path).values())[0]


def read_json_files_in_folder(filename):
    download_folder = os.path.join(os.path.dirname(filename), benedict.benedict(filename).get("folder"))
    return path_to_dict(download_folder)


def read_json_from_zip(filename, pattern: None):
    """
    Traverses a zip archive, parse the first *.json file and return its parsed
    content
    """
    import zipfile
    import json
    import re
    with zipfile.ZipFile(filename, "r") as z:
        for filename in z.namelist():
            if pattern and re.match(pattern, filename):
                with z.open(filename) as f:
                    return json.loads(f.read())



from beancount.core import data
from beancount.core import amount
from beancount.ingest import importer, cache
import re
import os
from .utils import map_dict, parse_amount, parse_date
from collections import namedtuple
import dateutil.parser
from beancount.core.number import D
from beancount.core import flags
import datetime
from beancount.loader import load_file
import itertools
from .logger import logger
from .utils import map_dict, parse_date, critical_error
import benedict


class Importer(importer.ImporterProtocol):

    _existing_duplication_keys = None

    _name = None

    def __init__(
        self,
        name: str = None,
        file_pattern: str = None,
        target_account: str = None,
        duplication_key=None,
        mapping_transactions=None,
        mapping_transaction=None,
        mapping_balances=None,
        mapping_balance=None,
        hook_prepare_data: callable = None
    ):
        self._name = name
        self._file_pattern = file_pattern
        self._target_account = target_account
        self._duplication_key = duplication_key
        self._mapping_transactions = mapping_transactions
        self._mapping_transaction = mapping_transaction
        self._mapping_balances = mapping_balances
        self._mapping_balance = mapping_balance
        self._hook_prepare_data = hook_prepare_data

    def name(self):
        if self._name:
            return self._name
        return "Generic yabci Importer"

    def identifier(self, filename) -> str:
        """
        A unique identifier for this importer+config
        """
        return "%s-%s" % (self.name(), self.basename(filename))

    def basename(self, filename):
        return '{}'.format(os.path.basename(filename))

    def identify(self, file: cache._FileMemo):
        if self._file_pattern:

            filename = file.name.lower()

            if not re.match(self._file_pattern, filename):
                logger.debug("Filename %s does not match pattern (%s)" % (filename, self._file_pattern))
                return False

        return True

    def file_name(self, file):
        return self.basename(file.name)

    def file_account(self, file):
        return self._target_account

    def file_date(self, file):
        # TODO is that suitable?
        return datetime.datetime.now().isoformat(timespec="seconds")

    def new_metadata(self, filename, counter, additional = {}):
        return {
            **data.new_metadata(filename, next(counter)),
            **(additional or {}),
        }

    def is_transaction_duplicate(self, transaction: data.Transaction, existing_entries: list) -> bool:
        """
        Check if a transaction exists already. Use a custom callable to
        determine, if two transactions represent the same (can return a certain
        id for instance)

        @return True if transaction exists, False otherwise.
        """
        callback = self._duplication_key

        # if it is a simple string, we treat it as array index (with dot syntax,
        # sth like "meta.customid")
        if isinstance(callback, str):
            key = callback
            callback = lambda transaction: benedict.benedict(transaction._asdict()).get(key)

        # no config for duplication detection -> return
        if not callback:
            return False

        # find the key for duplicate detection for the given transaction
        duplication_key = callback(transaction)

        # if the needle transaction does not have a duplication key, directly
        # return
        if not duplication_key:
            return False

        # now compute duplication keys for all existing entries (if not
        # previously happened)
        if self._existing_duplication_keys is None:
            existing_entries = existing_entries or []
            self._existing_duplication_keys = [callback(entry) for entry in existing_entries]

        return (duplication_key in self._existing_duplication_keys)

    def map_transaction(self, raw: dict) -> data.Transaction:
        """
        Map a raw data dict representing a transaction into a beancount
        Transaction object, considering the mapping config
        """
        try:
            if not self._mapping_transaction:
                raise Exception("Mapping for single transaction is not defined (key 'mapping_transaction')")

            mapped = map_dict(self._mapping_transaction, raw)

            # create a beancount object from the mapped values, with sensible
            # defaults
            transaction = data.Transaction(
                mapped.get("meta", data.EMPTY_SET),
                parse_date(mapped.get("date")),
                mapped.get("flag", flags.FLAG_OKAY),
                mapped.get("payee", None),
                mapped.get("narration", None),
                mapped.get("tags", data.EMPTY_SET),
                mapped.get("links", data.EMPTY_SET),
                [],
            )

        except Exception as e:
            raise Exception("Could not map transaction: %s\n\nused mapping config: %s\nraw data: %s" % (e, self._mapping_transaction, raw))

        for posting_data in mapped.get("postings", []):

            try:
                transaction.postings.append(data.Posting(
                    posting_data.get("account", self._target_account),
                    parse_amount(posting_data.get("amount")),
                    posting_data.get("cost", None),
                    parse_amount(posting_data.get("price", None)),
                    posting_data.get("flag", None),
                    posting_data.get("meta", data.EMPTY_SET), # meta
                ))

            except Exception as e:
                raise critical_error("Could not map posting: %s\n\nraw data: %s" % (e, posting_data))

        return transaction

    def map_balance(self, raw) -> data.Balance:
        """
        Map a raw data dict representing a balance into a beancount
        Balance object, considering the mapping config
        """

        try:
            if not self._mapping_balance:
                raise Exception("Mapping for single balance is not defined (key 'mapping_balance')")

            mapped = map_dict(self._mapping_balance, raw)

            return data.Balance(
                mapped.get("meta", data.EMPTY_SET),
                parse_date(mapped.get("date")),
                mapped.get("account", self._target_account),
                parse_amount(mapped.get("amount")),
                mapped.get("tolerance", None),
                parse_amount(mapped.get("diff_amount", None)),
            )

        except Exception as e:
            raise Exception("Could not map balance: %s\n\nused mapping config: %s\nraw data: %s" % (e, self._mapping_balance, raw))

    def get_transactions(self, data_raw: dict):
        """
        Take the raw input (benedict) and extract the list of (raw) transactions
        """
        if not self._mapping_transactions:
            return []

        raw_transactions = map_dict(self._mapping_transactions, data_raw)

        if raw_transactions is None:
            raise Exception("Could not find transactions with the key %s\n\navailable keys: %s" % (self._mapping_transactions, ",".join(data_raw.keys())))

        for raw_transaction in raw_transactions:
            yield self.map_transaction(raw_transaction)

    def get_balances(self, data_raw):
        """
        Take the raw input (benedict) and extract the list of (raw) balances
        """
        if not self._mapping_balances:
            return []

        raw_balances = map_dict(self._mapping_balances, data_raw)

        if raw_balances is None:
            raise Exception("Could not find balances with the key %s\n\navailable keys: %s" % (self._mapping_balances, ",".join(data_raw.keys())))

        for raw_balance in raw_balances:
            yield self.map_balance(raw_balance)

    def extract(self, file, existing_entries=None):

        # default: input passed to benedict is just the name of the import file
        benedict_input = file.name

        # if a hook is given, execute it
        if self._hook_prepare_data:
            logger.info("running prepare_data hook")
            benedict_input = self._hook_prepare_data(benedict_input)

        # load it into a benedict
        data_raw = benedict.benedict(benedict_input)

        counter = itertools.count()
        new_entries = []

        for entry in self.get_transactions(data_raw):
            metadata = self.new_metadata(file.name, counter, entry.meta or {})

            transaction = data.Transaction(**{**entry._asdict(), "meta": metadata})

            if self.is_transaction_duplicate(transaction, existing_entries):
                transaction.meta["__duplicate__"] = True

            new_entries.append(transaction)

        for balance in self.get_balances(data_raw):
            metadata = self.new_metadata(file.name, counter, balance.meta or {})
            balance = data.Balance(**{**balance._asdict(), "meta": metadata})

            if self.is_transaction_duplicate(balance, existing_entries):
                balance.meta["__duplicate__"] = True

            new_entries.append(balance)

        # metadata = data.new_metadata(file.name, next(counter))
        # new_entries.append(data.Custom(metadata, datetime.date.today(), "fints-importdate", []))

        return(new_entries)



import os.path
import unittest
import yabci
from beancount.ingest import extract
import datetime
from beancount.core.data import Transaction, Posting
from beancount.core.amount import Amount
from decimal import Decimal
import subprocess


class TestQuery(unittest.TestCase):

    def setUp(self):
        self.maxDiff = None
        self.path = None

    def _run_bean_extract(self, *args) -> str:
        cmd = ["bean-extract", *args]
        result = subprocess.run(cmd, cwd=self.path, stdout=subprocess.PIPE)

        # remove first two lines, since the contain the absolute path as comment
        # (which differs on each machine)
        plain_imported = "\n".join(result.stdout.decode('utf-8').split("\n")[2:]).strip()

        return plain_imported

    def _readfile(self, filename) -> str:
        fullpath = os.path.join(self.path, filename)
        return open(fullpath, "r").read().strip()

    def test_example_csv_basic(self):

        self.path = os.path.dirname(__file__) + "/../examples/csv-basic/"

        imported = self._run_bean_extract("config-csv-basic.py", "foo-bank-sample.csv")
        expected = self._readfile("expected-output.beancount")

        self.assertEqual(expected, imported)

    def test_example_csv_full(self):

        self.path = os.path.dirname(__file__) + "/../examples/csv-full/"

        imported = self._run_bean_extract(
            "-e",
            "existing.beancount",
            "config-csv-full.py",
            "venmo-sample.csv",
        )

        expected = self._readfile("expected-output.beancount")

        self.assertEqual(expected, imported)

    def test_example_android_moneywallet(self):

        self.path = os.path.dirname(__file__) + "/../examples/android-moneywallet/"

        imported = self._run_bean_extract(
            "-e",
            "existing.beancount",
            "config-android-moneywallet.py",
            "database.json",
        )

        expected = self._readfile("expected-output.beancount")

        self.assertEqual(expected, imported)

    def test_example_readme(self):

        self.path = os.path.dirname(__file__) + "/../examples/readme/"

        imported = self._run_bean_extract(
            "config.py",
            "sample.csv",
        )

        expected = self._readfile("expected-output.beancount")

        self.assertEqual(expected, imported)



import logging
import sys


def get_logger(name) -> logging.Logger:
    logger = logging.getLogger(name)
    logger.setLevel(logging.DEBUG)

    handler = logging.StreamHandler(sys.stdout)
    handler.setFormatter(logging.Formatter(
        '[%(asctime)s][%(name)s.%(levelname)s] %(message)s',
        datefmt='%Y-%m-%d %H:%M:%S'
    ))

    logger.addHandler(handler)

    return logger


logger = get_logger('fints-import')



import unittest
from beancount.core import data
from src.importer import Importer
from beancount.ingest import cache
import datetime


class TestImporter(unittest.TestCase):

    def setUp(self):
        self.maxDiff = None

    def test_name(self):
        self.assertEqual(Importer(name="foo bar").name(), "foo bar")
        self.assertEqual(Importer().name(), "Generic yabci Importer")

    def test_identify(self):

        csv_file = cache._FileMemo("/home/bob/test.csv")
        json_file = cache._FileMemo("/home/bob/test.json")

        # no pattern
        importer = Importer()
        self.assertTrue(importer.identify(csv_file))
        self.assertTrue(importer.identify(json_file))

        # some pattern
        importer = Importer(file_pattern=".*\\.csv")
        self.assertTrue(importer.identify(csv_file))
        self.assertFalse(importer.identify(json_file))

    def test_is_duplicate(self):

        transaction = data.Transaction(
            {"id": "id_foo"},
            None,
            None,
            None,
            None,
            None,
            None,
            [],
        )

        transaction_bar = data.Transaction(
            {"id": "id_bar"},
            None,
            None,
            None,
            None,
            None,
            None,
            [],
        )

        self.assertTrue(Importer(duplication_key="meta.id").is_transaction_duplicate(transaction, [transaction]))
        self.assertFalse(Importer(duplication_key="meta.id").is_transaction_duplicate(transaction, [transaction_bar]))

    def test_map_transaction_no_mapping(self):
        importer = Importer()
        with self.assertRaises(Exception):
            importer.map_transaction({})

    def test_map_transaction(self):

        # case #1: no mapping passed
        importer = Importer()
        with self.assertRaises(Exception):
            importer.map_transaction({})

        # case #2: mapping passed
        importer = Importer(mapping_transaction={"date": "raw_date", "narration": "raw_desc"})

        transaction = importer.map_transaction({"raw_date": "2023-01-01T09:12:34", "raw_desc": "descriptino"})

        self.assertEqual(datetime.date(2023, 1, 1), transaction.date)
        self.assertIsInstance(transaction, data.Transaction)
        self.assertEqual("descriptino", transaction.narration)

    def test_get_transactions(self):
        importer = Importer(mapping_transactions="raw_transactions", mapping_transaction={"date": "raw_date"})

        # case #1: key not found in input data
        transactions = importer.get_transactions({"invalid": [{"raw_date": "2023-01-01T09:12:34"}]})
        with self.assertRaises(Exception):
            list(transactions)

        # case #2: found & mapped
        transactions = list(importer.get_transactions({"raw_transactions": [{"raw_date": "2023-01-01T09:12:34"}, {"raw_date": "2023-02-03T09:12:34"}]}))

        self.assertEqual(2, len(transactions))
        self.assertIsInstance(transactions[0], data.Transaction)
        self.assertEqual(datetime.date(2023, 1, 1), transactions[0].date)
        self.assertEqual(datetime.date(2023, 2, 3), transactions[1].date)

    def test_map_balance(self):

        # case #1: no mapping passed
        importer = Importer()
        with self.assertRaises(Exception):
            importer.map_balance({})

        # case #2: mapping passed
        importer = Importer(mapping_balance={"date": "raw_date"})

        balance = importer.map_balance({"raw_date": "2023-01-01T09:12:34"})

        self.assertIsInstance(balance, data.Balance)
        self.assertEqual(datetime.date(2023, 1, 1), balance.date)

    def test_get_balances(self):
        importer = Importer(mapping_balances="raw_balances", mapping_balance={"date": "raw_date"})

        # case #1: key not found in input data
        balances = importer.get_balances({"invalid": [{"raw_date": "2023-01-01T09:12:34"}]})
        with self.assertRaises(Exception):
            list(balances)

        # case #2: found & mapped
        balances = list(importer.get_balances({"raw_balances": [{"raw_date": "2023-01-01T09:12:34"}, {"raw_date": "2023-02-03T09:12:34"}]}))

        self.assertEqual(2, len(balances))
        self.assertIsInstance(balances[0], data.Balance)
        self.assertEqual(datetime.date(2023, 1, 1), balances[0].date)
        self.assertEqual(datetime.date(2023, 2, 3), balances[1].date)


